# Copyright (c) 2010 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest
from unittest import mock

from urllib.parse import urlparse

from swift.common.swob import Request, Response, HTTPUnauthorized
from swift.common.middleware import staticweb


meta_map = {
    'c1': {'status': 401},
    'c2': {},
    'c3': {'meta': {'web-index': 'index.html',
                    'web-listings': 't'}},
    'c3b': {'meta': {'web-index': 'index.html',
                     'web-listings': 't'}},
    'c4': {'meta': {'web-index': 'index.html',
                    'web-error': 'error.html',
                    'web-listings': 't',
                    'web-listings-css': 'listing.css',
                    'web-directory-type': 'text/dir'}},
    'c5': {'meta': {'web-index': 'index.html',
                    'web-error': 'error.html',
                    'web-listings': 't',
                    'web-listings-css': 'listing.css'}},
    'c6': {'meta': {'web-listings': 't',
                    'web-error': 'error.html'}},
    'c6b': {'meta': {'web-listings': 't',
                     'web-listings-label': 'foo'}},
    'c7': {'meta': {'web-listings': 'f',
                    'web-error': 'error.html'}},
    'c8': {'meta': {'web-error': 'error.html',
                    'web-listings': 't',
                    'web-listings-css':
                    'http://localhost/stylesheets/listing.css'}},
    'c9': {'meta': {'web-error': 'error.html',
                    'web-listings': 't',
                    'web-listings-css':
                    '/absolute/listing.css'}},
    'c10': {'meta': {'web-listings': 't'}},
    'c11': {'meta': {'web-index': 'index.html'}},
    'c11a': {'meta': {'web-index': 'index.html',
             'web-directory-type': 'text/directory'}},
    'c12': {'meta': {'web-index': 'index.html',
                     'web-error': 'error.html'}},
    'c13': {'meta': {'web-listings': 'f',
                     'web-listings-css': 'listing.css'}},
    'c14': {'meta': {'web-listings': 't'}},
}


def mock_get_container_info(env, app, swift_source='SW'):
    container = env['PATH_INFO'].rstrip('/').split('/')[3]
    container_info = meta_map[container]
    container_info.setdefault('status', 200)
    container_info.setdefault('read_acl', '.r:*')
    return container_info


class FakeApp(object):

    def __init__(self, status_headers_body_iter=None):
        self.calls = 0
        self.get_c4_called = False

    def __call__(self, env, start_response):
        self.calls += 1
        if 'swift.authorize' in env:
            resp = env['swift.authorize'](Request(env))
            if resp:
                return resp(env, start_response)
        if env['PATH_INFO'] == '/':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1':
            return Response(
                status='412 Precondition Failed')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a':
            return Response(status='401 Unauthorized')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c1':
            return Response(status='401 Unauthorized')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c2':
            return self.listing(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c2/one.txt':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c3':
            return self.listing(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c3/index.html':
            return Response(status='200 Ok', body='''
<html>
    <body>
        <h1>Test main index.html file.</h1>
        <p>Visit <a href="subdir">subdir</a>.</p>
        <p>Don't visit <a href="subdir2/">subdir2</a> because it doesn't really
           exist.</p>
        <p>Visit <a href="subdir3">subdir3</a>.</p>
        <p>Visit <a href="subdir3/subsubdir">subdir3/subsubdir</a>.</p>
    </body>
</html>
            ''')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c3b':
            return self.listing(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c3b/index.html':
            resp = Response(status='204 No Content')
            resp.app_iter = iter([])
            return resp(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c3/subdir':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c3/subdir/':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c3/subdir/index.html':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c3/subdir3/subsubdir':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c3/subdir3/subsubdir/':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c3/subdir3/subsubdir/index.html':
            return Response(status='200 Ok', body='index file')(env,
                                                                start_response)
        elif env['PATH_INFO'] == '/v1/a/c3/subdirx/':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c3/subdirx/index.html':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c3/subdiry/':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c3/subdiry/index.html':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c3/subdirz':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c3/subdirz/index.html':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c3/unknown':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c3/unknown/index.html':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c4':
            self.get_c4_called = True
            return self.listing(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c4/one.txt':
            return Response(
                status='200 Ok',
                headers={'x-object-meta-test': 'value'},
                body='1')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c4/two.txt':
            return Response(status='503 Service Unavailable')(env,
                                                              start_response)
        elif env['PATH_INFO'] == '/v1/a/c4/index.html':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c4/subdir/':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c4/subdir/index.html':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c4/unknown':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c4/unknown/index.html':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c4/404error.html':
            return Response(status='200 Ok', body='''
<html>
    <body style="background: #000000; color: #ffaaaa">
        <p>Chrome's 404 fancy-page sucks.</p>
    </body>
</html>
            '''.strip())(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c5':
            return self.listing(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c5/index.html':
            return Response(status='503 Service Unavailable')(env,
                                                              start_response)
        elif env['PATH_INFO'] == '/v1/a/c5/503error.html':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c5/unknown':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c5/unknown/index.html':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c5/404error.html':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c6':
            return self.listing(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c6b':
            return self.listing(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c6/subdir':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c6/401error.html':
            return Response(status='200 Ok', body='''
<html>
    <body style="background: #000000; color: #ffaaaa">
        <p>Hey, you're not authorized to see this!</p>
    </body>
</html>
            '''.strip())(env, start_response)
        elif env['PATH_INFO'] in ('/v1/a/c7', '/v1/a/c7/'):
            return self.listing(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c7/404error.html':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c7/401error.html':
            return Response(status='200 Ok', body='''
<html>
    <body style="background: #000000; color: #ffaaaa">
        <p>Hey, you're not authorized to see this!</p>
    </body>
</html>
            '''.strip())(env, start_response)
        elif env['PATH_INFO'] in ('/v1/a/c8', '/v1/a/c8/'):
            return self.listing(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c8/subdir/':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] in ('/v1/a/c9', '/v1/a/c9/'):
            return self.listing(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c9/subdir/':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] in ('/v1/a/c10', '/v1/a/c10/'):
            return self.listing(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c10/\xe2\x98\x83/':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c10/\xe2\x98\x83/\xe2\x98\x83/':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] in ('/v1/a/c11', '/v1/a/c11/'):
            return self.listing(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c11/subdir/':
            return Response(status='200 Ok', headers={
                'Content-Type': 'application/directory'})(
                    env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c11/subdir/index.html':
            return Response(status='200 Ok', body='''
<html>
    <body>
        <h2>c11 subdir index</h2>
    </body>
</html>
            '''.strip())(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c11/subdir2/':
            return Response(status='200 Ok', headers={'Content-Type':
                            'application/directory'})(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c11/subdir2/index.html':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] in ('/v1/a/c11a', '/v1/a/c11a/'):
            return self.listing(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c11a/subdir/':
            return Response(status='200 Ok', headers={'Content-Type':
                            'text/directory'})(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c11a/subdir/index.html':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c11a/subdir2/':
            return Response(status='200 Ok', headers={'Content-Type':
                            'application/directory'})(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c11a/subdir2/index.html':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c11a/subdir3/':
            return Response(status='200 Ok', headers={'Content-Type':
                            'not_a/directory'})(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c11a/subdir3/index.html':
            return Response(status='404 Not Found')(env, start_response)
        elif env['PATH_INFO'] == '/v1/a/c12/index.html':
            return Response(status='200 Ok', body='index file')(env,
                                                                start_response)
        elif env['PATH_INFO'] == '/v1/a/c12/200error.html':
            return Response(status='200 Ok', body='error file')(env,
                                                                start_response)
        elif env['PATH_INFO'] == '/v1/a/c14':
            return self.listing(env, start_response)
        else:
            raise Exception('Unknown path %r' % env['PATH_INFO'])

    def listing(self, env, start_response):
        headers = {'x-container-read': '.r:*'}
        if ((env['PATH_INFO'] in (
                '/v1/a/c3', '/v1/a/c4', '/v1/a/c8', '/v1/a/c9'))
            and (env['QUERY_STRING'] ==
                 'delimiter=/&prefix=subdir/')):
            headers.update({'X-Container-Object-Count': '12',
                            'X-Container-Bytes-Used': '73763',
                            'X-Container-Read': '.r:*',
                            'Content-Type': 'application/json; charset=utf-8'})
            body = '''
                [{"name":"subdir/1.txt",
                  "hash":"5f595114a4b3077edfac792c61ca4fe4", "bytes":20,
                  "content_type":"text/plain",
                  "last_modified":"2011-03-24T04:27:52.709100"},
                 {"name":"subdir/2.txt",
                  "hash":"c85c1dcd19cf5cbac84e6043c31bb63e", "bytes":20,
                  "content_type":"text/plain",
                  "last_modified":"2011-03-24T04:27:52.734140"},
                 {"subdir":"subdir3/subsubdir/"}]
            '''.strip()
        elif env['PATH_INFO'] == '/v1/a/c3' and env['QUERY_STRING'] == \
                'delimiter=/&prefix=subdiry/':
            headers.update({'X-Container-Object-Count': '12',
                            'X-Container-Bytes-Used': '73763',
                            'X-Container-Read': '.r:*',
                            'Content-Type': 'application/json; charset=utf-8'})
            body = '[]'
        elif env['PATH_INFO'] == '/v1/a/c3' and env['QUERY_STRING'] == \
                'limit=1&delimiter=/&prefix=subdirz/':
            headers.update({'X-Container-Object-Count': '12',
                            'X-Container-Bytes-Used': '73763',
                            'X-Container-Read': '.r:*',
                            'Content-Type': 'application/json; charset=utf-8'})
            body = '''
                [{"name":"subdirz/1.txt",
                  "hash":"5f595114a4b3077edfac792c61ca4fe4", "bytes":20,
                  "content_type":"text/plain",
                  "last_modified":"2011-03-24T04:27:52.709100"}]
            '''.strip()
        elif env['PATH_INFO'] == '/v1/a/c6' and env['QUERY_STRING'] == \
                'limit=1&delimiter=/&prefix=subdir/':
            headers.update({'X-Container-Object-Count': '12',
                            'X-Container-Bytes-Used': '73763',
                            'X-Container-Read': '.r:*',
                            'X-Container-Web-Listings': 't',
                            'Content-Type': 'application/json; charset=utf-8'})
            body = '''
                [{"name":"subdir/1.txt",
                  "hash":"5f595114a4b3077edfac792c61ca4fe4", "bytes":20,
                  "content_type":"text/plain",
                  "last_modified":"2011-03-24T04:27:52.709100"}]
            '''.strip()
        elif env['PATH_INFO'] == '/v1/a/c10' and (
                env['QUERY_STRING'] ==
                'delimiter=/&prefix=%E2%98%83/' or
                env['QUERY_STRING'] ==
                'delimiter=/&prefix=%E2%98%83/%E2%98%83/'):
            headers.update({'X-Container-Object-Count': '12',
                            'X-Container-Bytes-Used': '73763',
                            'X-Container-Read': '.r:*',
                            'X-Container-Web-Listings': 't',
                            'Content-Type': 'application/json; charset=utf-8'})
            body = '''
                [{"name":"\u2603/\u2603/one.txt",
                  "hash":"73f1dd69bacbf0847cc9cffa3c6b23a1", "bytes":22,
                  "content_type":"text/plain",
                  "last_modified":"2011-03-24T04:27:52.709100"},
                 {"subdir":"\u2603/\u2603/"}]
            '''.strip()
        elif env['PATH_INFO'] == '/v1/a/c14' and env['QUERY_STRING'] == \
                'delimiter=/':
            headers.update({'X-Container-Object-Count': '0',
                            'X-Container-Bytes-Used': '0',
                            'X-Container-Read': '.r:*',
                            'Content-Type': 'application/json; charset=utf-8'})
            body = '[]'
        elif 'prefix=' in env['QUERY_STRING']:
            return Response(status='204 No Content')(env, start_response)
        else:
            headers.update({'X-Container-Object-Count': '12',
                            'X-Container-Bytes-Used': '73763',
                            'Content-Type': 'application/json; charset=utf-8'})
            body = '''
                [{"name":"401error.html",
                  "hash":"893f8d80692a4d3875b45be8f152ad18", "bytes":110,
                  "content_type":"text/html",
                  "last_modified":"2011-03-24T04:27:52.713710"},
                 {"name":"404error.html",
                  "hash":"62dcec9c34ed2b347d94e6ca707aff8c", "bytes":130,
                  "content_type":"text/html",
                  "last_modified":"2011-03-24T04:27:52.720850"},
                 {"name":"index.html",
                  "hash":"8b469f2ca117668a5131fe9ee0815421", "bytes":347,
                  "content_type":"text/html",
                  "last_modified":"2011-03-24T04:27:52.683590"},
                 {"name":"listing.css",
                  "hash":"7eab5d169f3fcd06a08c130fa10c5236", "bytes":17,
                  "content_type":"text/css",
                  "last_modified":"2011-03-24T04:27:52.721610"},
                 {"name":"one.txt", "hash":"73f1dd69bacbf0847cc9cffa3c6b23a1",
                  "bytes":22, "content_type":"text/plain",
                  "last_modified":"2011-03-24T04:27:52.722270"},
                 {"name":"subdir/1.txt",
                  "hash":"5f595114a4b3077edfac792c61ca4fe4", "bytes":20,
                  "content_type":"text/plain",
                  "last_modified":"2011-03-24T04:27:52.709100"},
                 {"name":"subdir/2.txt",
                  "hash":"c85c1dcd19cf5cbac84e6043c31bb63e", "bytes":20,
                  "content_type":"text/plain",
                  "last_modified":"2011-03-24T04:27:52.734140"},
                 {"name":"subdir/\u2603.txt",
                  "hash":"7337d028c093130898d937c319cc9865", "bytes":72981,
                  "content_type":"text/plain",
                  "last_modified":"2011-03-24T04:27:52.735460"},
                 {"name":"subdir2", "hash":"d41d8cd98f00b204e9800998ecf8427e",
                  "bytes":0, "content_type":"text/directory",
                  "last_modified":"2011-03-24T04:27:52.676690"},
                 {"name":"subdir3/subsubdir/index.html",
                  "hash":"04eea67110f883b1a5c97eb44ccad08c", "bytes":72,
                  "content_type":"text/html",
                  "last_modified":"2011-03-24T04:27:52.751260"},
                 {"name":"two.txt", "hash":"10abb84c63a5cff379fdfd6385918833",
                  "bytes":22, "content_type":"text/plain",
                  "last_modified":"2011-03-24T04:27:52.825110"},
                 {"name":"\u2603/\u2603/one.txt",
                  "hash":"73f1dd69bacbf0847cc9cffa3c6b23a1", "bytes":22,
                  "content_type":"text/plain",
                  "last_modified":"2011-03-24T04:27:52.935560"}]
            '''.strip()
        return Response(status='200 Ok', headers=headers,
                        body=body)(env, start_response)


class FakeAuthFilter(object):

    def __init__(self, app, deny_objects=False, deny_listing=False):
        self.app = app
        self.deny_objects = deny_objects
        self.deny_listing = deny_listing

    def authorize(self, req):
        path_parts = req.path.strip('/').split('/')
        if ((self.deny_objects and len(path_parts) > 3)
                or (self.deny_listing and len(path_parts) == 3)):
            return HTTPUnauthorized()

    def __call__(self, env, start_response):
        env['swift.authorize'] = self.authorize
        return self.app(env, start_response)


class TestStaticWeb(unittest.TestCase):

    def setUp(self):
        self.app = FakeApp()
        self.test_staticweb = FakeAuthFilter(
            staticweb.filter_factory({})(self.app))
        self._orig_get_container_info = staticweb.get_container_info
        staticweb.get_container_info = mock_get_container_info

    def tearDown(self):
        staticweb.get_container_info = self._orig_get_container_info

    def test_app_set(self):
        app = FakeApp()
        sw = staticweb.filter_factory({})(app)
        self.assertEqual(sw.app, app)

    def test_conf_set(self):
        conf = {'blah': 1}
        sw = staticweb.filter_factory(conf)(FakeApp())
        self.assertEqual(sw.conf, conf)

    def test_root(self):
        resp = Request.blank('/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 404)

    def test_version(self):
        resp = Request.blank('/v1').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 412)

    def test_account(self):
        resp = Request.blank('/v1/a').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 401)

    def test_container1(self):
        resp = Request.blank('/v1/a/c1').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 401)

    def test_container1_web_mode_explicitly_off(self):
        resp = Request.blank('/v1/a/c1',
                             headers={'x-web-mode': 'false'}).get_response(
                                 self.test_staticweb)
        self.assertEqual(resp.status_int, 401)

    def test_container1_web_mode_explicitly_on(self):
        resp = Request.blank('/v1/a/c1',
                             headers={'x-web-mode': 'true'}).get_response(
                                 self.test_staticweb)
        self.assertEqual(resp.status_int, 404)

    def test_container2(self):
        resp = Request.blank('/v1/a/c2').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertEqual(resp.content_type, 'application/json')
        self.assertEqual(len(json.loads(resp.body)),
                         int(resp.headers['x-container-object-count']))

    def test_container2_web_mode_explicitly_off(self):
        resp = Request.blank(
            '/v1/a/c2',
            headers={'x-web-mode': 'false'}).get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertEqual(resp.content_type, 'application/json')
        self.assertEqual(len(json.loads(resp.body)),
                         int(resp.headers['x-container-object-count']))

    def test_container2_web_mode_explicitly_on(self):
        resp = Request.blank(
            '/v1/a/c2',
            headers={'x-web-mode': 'true'}).get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 404)

    def test_container2onetxt(self):
        resp = Request.blank(
            '/v1/a/c2/one.txt').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 404)

    def test_container2json(self):
        resp = Request.blank(
            '/v1/a/c2').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertEqual(resp.content_type, 'application/json')
        self.assertEqual(len(json.loads(resp.body)),
                         int(resp.headers['x-container-object-count']))

    def test_container2json_web_mode_explicitly_off(self):
        resp = Request.blank(
            '/v1/a/c2',
            headers={'x-web-mode': 'false'}).get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertEqual(resp.content_type, 'application/json')
        self.assertEqual(len(json.loads(resp.body)),
                         int(resp.headers['x-container-object-count']))

    def test_container2json_web_mode_explicitly_on(self):
        resp = Request.blank(
            '/v1/a/c2',
            headers={'x-web-mode': 'true'}).get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 404)

    def test_container3(self):
        resp = Request.blank('/v1/a/c3').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 301)
        self.assertEqual(resp.headers['location'],
                         'http://localhost/v1/a/c3/')

    def test_container3indexhtml(self):
        resp = Request.blank('/v1/a/c3/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(b'Test main index.html file.', resp.body)
        self.assertNotIn('X-Backend-Content-Generator', resp.headers)

    def test_container3subsubdir(self):
        resp = Request.blank(
            '/v1/a/c3/subdir3/subsubdir').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 301)

    def test_container3subsubdircontents(self):
        resp = Request.blank(
            '/v1/a/c3/subdir3/subsubdir/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertEqual(resp.body, b'index file')

    def test_container3subdir(self):
        resp = Request.blank(
            '/v1/a/c3/subdir/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(b'Listing of /v1/a/c3/subdir/', resp.body)
        self.assertIn(b'</style>', resp.body)
        self.assertNotIn(b'<link', resp.body)
        self.assertNotIn(b'listing.css', resp.body)

    def test_container3subdirx(self):
        resp = Request.blank(
            '/v1/a/c3/subdirx/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 404)

    def test_container3subdiry(self):
        resp = Request.blank(
            '/v1/a/c3/subdiry/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 404)

    def test_container3subdirz(self):
        resp = Request.blank(
            '/v1/a/c3/subdirz').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 301)

    def test_container3unknown(self):
        resp = Request.blank(
            '/v1/a/c3/unknown').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 404)
        self.assertNotIn(b"Chrome's 404 fancy-page sucks.", resp.body)

    def test_container3bindexhtml(self):
        resp = Request.blank('/v1/a/c3b/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 204)
        self.assertEqual(resp.body, b'')

    def test_container4indexhtml(self):
        resp = Request.blank('/v1/a/c4/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(b'Listing of /v1/a/c4/', resp.body)
        self.assertIn(b'href="listing.css"', resp.body)
        self.assertIn('X-Backend-Content-Generator', resp.headers)
        self.assertEqual(resp.headers['X-Backend-Content-Generator'],
                         'staticweb')

    def test_container4indexhtmlauthed(self):
        # anonymous access gets staticweb
        resp = Request.blank('/v1/a/c4').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 301)

        # authed access doesn't (by default)
        resp = Request.blank(
            '/v1/a/c4',
            environ={'REMOTE_USER': 'authed'}).get_response(
                self.test_staticweb)
        self.assertEqual(resp.status_int, 200)

        # it can opt-in, though!
        resp = Request.blank(
            '/v1/a/c4', headers={'x-web-mode': 't'},
            environ={'REMOTE_USER': 'authed'}).get_response(
                self.test_staticweb)
        self.assertEqual(resp.status_int, 301)

        # and there's an exclusion for authed-via-tempurl
        resp = Request.blank(
            '/v1/a/c4',
            environ={'REMOTE_USER': '.wsgi.tempurl'}).get_response(
                self.test_staticweb)
        self.assertEqual(resp.status_int, 301)

    def test_container4tempurl(self):
        parts = [
            'temp_url_prefix=subdir/',
            'temp_url_sig=the-sig',
            'temp_url_expires=2024-12-31T00:00:00'
        ]

        resp = Request.blank(
            '/v1/a/c4/subdir/?' + '&'.join(parts),
            environ={'REMOTE_USER': '.wsgi.tempurl'},
        ).get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(b'Listing of /v1/a/c4/subdir/', resp.body)
        self.assertIn(b'<a href="2.txt?temp_url_prefix=subdir/&amp;'
                      b'temp_url_expires=2024-12-31T00%3A00%3A00&amp;'
                      b'temp_url_sig=the-sig">2.txt</a>', resp.body)

        parts.append('temp_url_ip_range=127.0.0.1')
        resp = Request.blank(
            '/v1/a/c4/subdir/?' + '&'.join(parts),
            environ={'REMOTE_USER': '.wsgi.tempurl'},
        ).get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(b'Listing of /v1/a/c4/subdir/', resp.body)
        self.assertIn(b'<a href="2.txt?temp_url_prefix=subdir/&amp;'
                      b'temp_url_expires=2024-12-31T00%3A00%3A00&amp;'
                      b'temp_url_sig=the-sig&amp;temp_url_ip_range='
                      b'127.0.0.1">2.txt</a>', resp.body)

        parts.append('inline')
        resp = Request.blank(
            '/v1/a/c4/subdir/?' + '&'.join(parts),
            environ={'REMOTE_USER': '.wsgi.tempurl'},
        ).get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(b'Listing of /v1/a/c4/subdir/', resp.body)
        self.assertIn(b'<a href="2.txt?temp_url_prefix=subdir/&amp;'
                      b'temp_url_expires=2024-12-31T00%3A00%3A00&amp;'
                      b'temp_url_sig=the-sig&amp;temp_url_ip_range='
                      b'127.0.0.1&amp;inline">2.txt</a>', resp.body)

        # no prefix => you get normal links (which will almost certainly 401)
        resp = Request.blank(
            '/v1/a/c4/subdir/?' + '&'.join(parts[1:]),
            environ={'REMOTE_USER': '.wsgi.tempurl'},
        ).get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(b'Listing of /v1/a/c4/subdir/', resp.body)
        self.assertIn(b'<a href="2.txt">2.txt</a>', resp.body)

    def test_container4unknown(self):
        resp = Request.blank(
            '/v1/a/c4/unknown').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 404)
        self.assertIn(b"Chrome's 404 fancy-page sucks.", resp.body)

    def test_container4subdir(self):
        resp = Request.blank(
            '/v1/a/c4/subdir/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(b'Listing of /v1/a/c4/subdir/', resp.body)
        self.assertNotIn(b'</style>', resp.body)
        self.assertIn(b'<link', resp.body)
        self.assertIn(b'href="../listing.css"', resp.body)
        self.assertEqual(resp.headers['content-type'],
                         'text/html; charset=UTF-8')

    def test_container4onetxt(self):
        resp = Request.blank(
            '/v1/a/c4/one.txt').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)

    def test_container4twotxt(self):
        resp = Request.blank(
            '/v1/a/c4/two.txt').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 503)

    def test_container5indexhtml(self):
        resp = Request.blank('/v1/a/c5/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 503)

    def test_container5unknown(self):
        resp = Request.blank(
            '/v1/a/c5/unknown').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 404)
        self.assertNotIn(b"Chrome's 404 fancy-page sucks.", resp.body)

    def test_container6subdir(self):
        resp = Request.blank(
            '/v1/a/c6/subdir').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 301)

    def test_container6listing(self):
        # container6 has web-listings = t, web-error=error.html
        resp = Request.blank('/v1/a/c6/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)

        # expect custom 401 if request is not auth'd for listing but is auth'd
        # to GET objects
        test_staticweb = FakeAuthFilter(
            staticweb.filter_factory({})(self.app), deny_listing=True)
        resp = Request.blank('/v1/a/c6/').get_response(test_staticweb)
        self.assertEqual(resp.status_int, 401)
        self.assertIn(b"Hey, you're not authorized to see this!", resp.body)

        # expect default 401 if request is not auth'd for listing or object GET
        test_staticweb = FakeAuthFilter(
            staticweb.filter_factory({})(self.app), deny_listing=True,
            deny_objects=True)
        resp = Request.blank('/v1/a/c6/').get_response(test_staticweb)
        self.assertEqual(resp.status_int, 401)
        self.assertNotIn(b"Hey, you're not authorized to see this!", resp.body)

    def test_container6blisting(self):
        label = 'Listing of {0}/'.format(
            meta_map['c6b']['meta']['web-listings-label'])
        resp = Request.blank('/v1/a/c6b/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(label.encode('utf-8'), resp.body)

    def test_container7listing(self):
        # container7 has web-listings = f, web-error=error.html
        resp = Request.blank('/v1/a/c7/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 404)
        self.assertIn(b"Web Listing Disabled", resp.body)

        # expect 301 if auth'd but no trailing '/'
        resp = Request.blank('/v1/a/c7').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 301)

        # expect default 401 if request is not auth'd and no trailing '/'
        test_staticweb = FakeAuthFilter(
            staticweb.filter_factory({})(self.app), deny_listing=True,
            deny_objects=True)
        resp = Request.blank('/v1/a/c7').get_response(test_staticweb)
        self.assertEqual(resp.status_int, 401)
        self.assertNotIn(b"Hey, you're not authorized to see this!", resp.body)

        # expect custom 401 if request is not auth'd for listing
        test_staticweb = FakeAuthFilter(
            staticweb.filter_factory({})(self.app), deny_listing=True)
        resp = Request.blank('/v1/a/c7/').get_response(test_staticweb)
        self.assertEqual(resp.status_int, 401)
        self.assertIn(b"Hey, you're not authorized to see this!", resp.body)

        # expect default 401 if request is not auth'd for listing or object GET
        test_staticweb = FakeAuthFilter(
            staticweb.filter_factory({})(self.app), deny_listing=True,
            deny_objects=True)
        resp = Request.blank('/v1/a/c7/').get_response(test_staticweb)
        self.assertEqual(resp.status_int, 401)
        self.assertNotIn(b"Hey, you're not authorized to see this!", resp.body)

    def test_container8listingcss(self):
        resp = Request.blank(
            '/v1/a/c8/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(b'Listing of /v1/a/c8/', resp.body)
        self.assertIn(b'<link', resp.body)
        self.assertIn(b'href="http://localhost/stylesheets/listing.css"',
                      resp.body)

    def test_container8subdirlistingcss(self):
        resp = Request.blank(
            '/v1/a/c8/subdir/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(b'Listing of /v1/a/c8/subdir/', resp.body)
        self.assertIn(b'<link', resp.body)
        self.assertIn(b'href="http://localhost/stylesheets/listing.css"',
                      resp.body)

    def test_container9listingcss(self):
        resp = Request.blank(
            '/v1/a/c9/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(b'Listing of /v1/a/c9/', resp.body)
        self.assertIn(b'<link', resp.body)
        self.assertIn(b'href="/absolute/listing.css"', resp.body)

    def test_container9subdirlistingcss(self):
        resp = Request.blank(
            '/v1/a/c9/subdir/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(b'Listing of /v1/a/c9/subdir/', resp.body)
        self.assertIn(b'<link', resp.body)
        self.assertIn(b'href="/absolute/listing.css"', resp.body)

    def test_container10unicodesubdirlisting(self):
        resp = Request.blank(
            '/v1/a/c10/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(b'Listing of /v1/a/c10/', resp.body)
        resp = Request.blank(
            '/v1/a/c10/\xe2\x98\x83/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(b'Listing of /v1/a/c10/\xe2\x98\x83/', resp.body)
        resp = Request.blank(
            '/v1/a/c10/\xe2\x98\x83/\xe2\x98\x83/'
        ).get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(
            b'Listing of /v1/a/c10/\xe2\x98\x83/\xe2\x98\x83/', resp.body)

    def test_container11subdirmarkerobjectindex(self):
        resp = Request.blank('/v1/a/c11/subdir/').get_response(
            self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(b'<h2>c11 subdir index</h2>', resp.body)

    def test_container11subdirmarkermatchdirtype(self):
        resp = Request.blank('/v1/a/c11a/subdir/').get_response(
            self.test_staticweb)
        self.assertEqual(resp.status_int, 404)
        self.assertIn(b'Index File Not Found', resp.body)

    def test_container11subdirmarkeraltdirtype(self):
        resp = Request.blank('/v1/a/c11a/subdir2/').get_response(
            self.test_staticweb)
        self.assertEqual(resp.status_int, 200)

    def test_container11subdirmarkerinvaliddirtype(self):
        resp = Request.blank('/v1/a/c11a/subdir3/').get_response(
            self.test_staticweb)
        self.assertEqual(resp.status_int, 200)

    def test_container12unredirectedrequest(self):
        resp = Request.blank('/v1/a/c12/').get_response(
            self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(b'index file', resp.body)

    def test_container13empty(self):
        resp = Request.blank(
            '/v1/a/c14/').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertIn(b'Listing of /v1/a/c14/', resp.body)
        self.assertIn(b'</style>', resp.body)
        self.assertNotIn(b'<link', resp.body)
        self.assertNotIn(b'listing.css', resp.body)
        self.assertNotIn(b'<td', resp.body)

    def test_container_404_has_css(self):
        resp = Request.blank('/v1/a/c13/').get_response(
            self.test_staticweb)
        self.assertEqual(resp.status_int, 404)
        self.assertIn(b'listing.css', resp.body)

    def test_container_404_has_no_css(self):
        resp = Request.blank('/v1/a/c7/').get_response(
            self.test_staticweb)
        self.assertEqual(resp.status_int, 404)
        self.assertNotIn(b'listing.css', resp.body)
        self.assertIn(b'<style', resp.body)

    def test_subrequest_once_if_possible(self):
        resp = Request.blank(
            '/v1/a/c4/one.txt').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)
        self.assertEqual(resp.headers['x-object-meta-test'], 'value')
        self.assertEqual(resp.body, b'1')
        self.assertEqual(self.app.calls, 1)

    def test_no_auth_middleware(self):
        resp = Request.blank('/v1/a/c3').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 301)
        # Test without an authentication middleware before staticweb
        # This is no longer handled by staticweb middleware, thus not returning
        # a 301 redirect
        self.test_staticweb = staticweb.filter_factory({})(self.app)
        resp = Request.blank('/v1/a/c3').get_response(self.test_staticweb)
        self.assertEqual(resp.status_int, 200)

    def test_subrequest_not_override_auth(self):
        app_call = \
            'swift.common.middleware.staticweb._StaticWebContext._app_call'
        orig_app_call = staticweb._StaticWebContext._app_call
        _fail = self.fail

        def hook_app_call(self, env):
            if 'swift.authorize_override' in env:
                _fail('staticweb must not create authorize info by itself')
            return orig_app_call(self, env)

        with mock.patch(app_call, hook_app_call):
            # testing for _listing container
            resp = Request.blank('/v1/a/c4/').get_response(self.test_staticweb)
            self.assertEqual(resp.status_int, 200)  # sanity

            # testing for _listing object subdir
            resp = Request.blank(
                '/v1/a/c4/unknown').get_response(self.test_staticweb)
            self.assertEqual(resp.status_int, 404)

            # testing for _error_response
            resp = Request.blank('/v1/a/c5/').get_response(self.test_staticweb)
            self.assertEqual(resp.status_int, 503)  # sanity


class TestStaticWebUrlBase(unittest.TestCase):

    def setUp(self):
        self.app = FakeApp()
        self._orig_get_container_info = staticweb.get_container_info
        staticweb.get_container_info = mock_get_container_info

    def tearDown(self):
        staticweb.get_container_info = self._orig_get_container_info

    def test_container3subdirz_scheme(self):
        path = '/v1/a/c3/subdirz'
        scheme = 'https'
        test_staticweb = FakeAuthFilter(
            staticweb.filter_factory({'url_base': 'https://'})(self.app))
        resp = Request.blank(path).get_response(test_staticweb)
        self.assertEqual(resp.status_int, 301)
        parsed = urlparse(resp.location)
        self.assertEqual(parsed.scheme, scheme)
        # We omit comparing netloc here, because swob is free to add port.
        self.assertEqual(parsed.path, path + '/')

    def test_container3subdirz_host(self):
        path = '/v1/a/c3/subdirz'
        netloc = 'example.com'
        test_staticweb = FakeAuthFilter(
            staticweb.filter_factory({
                'url_base': '//%s' % (netloc,)})(self.app))
        resp = Request.blank(path).get_response(test_staticweb)
        self.assertEqual(resp.status_int, 301)
        parsed = urlparse(resp.location)
        # We compare scheme with the default. This may change, but unlikely.
        self.assertEqual(parsed.scheme, 'http')
        self.assertEqual(parsed.netloc, netloc)
        self.assertEqual(parsed.path, path + '/')

    def test_container3subdirz_both(self):
        path = '/v1/a/c3/subdirz'
        scheme = 'http'
        netloc = 'example.com'
        test_staticweb = FakeAuthFilter(
            staticweb.filter_factory({
                'url_base': 'http://example.com'})(self.app))
        resp = Request.blank(path).get_response(test_staticweb)
        self.assertEqual(resp.status_int, 301)
        parsed = urlparse(resp.location)
        self.assertEqual(parsed.scheme, scheme)
        self.assertEqual(parsed.netloc, netloc)
        self.assertEqual(parsed.path, path + '/')


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2024 NVIDIA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import anycrc
except ImportError:
    anycrc = None
import binascii
import ctypes
import ctypes.util
import errno
import socket
import struct
import zlib


# See if anycrc is available...
if anycrc:
    crc32c_anycrc = anycrc.Model('CRC32C').calc
    crc64nvme_anycrc = anycrc.Model('CRC64-NVME').calc
else:
    crc32c_anycrc = None
    crc64nvme_anycrc = None


def find_isal():
    # If isal is available system-wide, great!
    isal_lib = ctypes.util.find_library('isal')
    if isal_lib is None:
        # py38+: Hopefully pyeclib was installed from a manylinux wheel
        # with isal baked in?
        try:
            import pyeclib  # noqa
            from importlib.metadata import \
                files as pkg_files, PackageNotFoundError  # py38+
        except ImportError:
            pass
        else:
            # Assume busted installs won't have it
            try:
                pyeclib_files = pkg_files('pyeclib')
                if pyeclib_files is None:
                    # Have a dist-info, but no RECORD file??
                    pyeclib_files = []
            except PackageNotFoundError:
                # Could import pyeclib, but no dist-info directory??
                pyeclib_files = []
            isal_libs = [f for f in pyeclib_files
                         if f.name.startswith("libisal")]
            if len(isal_libs) == 1:
                isal_lib = isal_libs[0].locate()
    return ctypes.CDLL(isal_lib) if isal_lib else None


isal = find_isal()

if hasattr(isal, 'crc32_iscsi'):  # isa-l >= 2.16
    isal.crc32_iscsi.argtypes = [ctypes.c_char_p, ctypes.c_int, ctypes.c_uint]
    isal.crc32_iscsi.restype = ctypes.c_uint

    def crc32c_isal(data, value=0):
        result = isal.crc32_iscsi(
            data,
            len(data),
            value ^ 0xffff_ffff,
        )
        # for some reason, despite us specifying that restype is uint,
        # it can come back signed??
        return (result & 0xffff_ffff) ^ 0xffff_ffff
else:
    crc32c_isal = None

if hasattr(isal, 'crc64_rocksoft_refl'):  # isa-l >= 2.31.0
    isal.crc64_rocksoft_refl.argtypes = [
        ctypes.c_uint64, ctypes.c_char_p, ctypes.c_uint64]
    isal.crc64_rocksoft_refl.restype = ctypes.c_uint64

    def crc64nvme_isal(data, value=0):
        return isal.crc64_rocksoft_refl(
            value,
            data,
            len(data),
        )
else:
    crc64nvme_isal = None


# The kernel may also provide crc32c
AF_ALG = getattr(socket, 'AF_ALG', 38)
try:
    _sock = socket.socket(AF_ALG, socket.SOCK_SEQPACKET)
    _sock.bind(("hash", "crc32c"))
except OSError as e:
    if e.errno == errno.ENOENT:
        # could create socket, but crc32c is unknown
        _sock.close()
    elif e.errno != errno.EAFNOSUPPORT:
        raise
    crc32c_kern = None
else:
    def crc32c_kern(data, value=0):
        crc32c_sock = socket.socket(AF_ALG, socket.SOCK_SEQPACKET)
        try:
            crc32c_sock.bind(("hash", "crc32c"))
            crc32c_sock.setsockopt(
                socket.SOL_ALG,
                socket.ALG_SET_KEY,
                struct.pack("I", value ^ 0xffff_ffff))
            sock, _ = crc32c_sock.accept()
            try:
                sock.sendall(data)
                return struct.unpack("I", sock.recv(4))[0]
            finally:
                sock.close()
        finally:
            crc32c_sock.close()


def _select_crc32c_impl():
    # Use the best implementation available.
    # On various hardware we've seen
    #
    #  CPU           |   ISA-L   |  Kernel
    # ---------------+-----------+----------
    # Intel N100     |  ~9GB/s   | ~3.5GB/s
    # ARM Cortex-A55 |  ~2.5GB/s | ~0.4GB/s
    # Intel 11850H   |  ~7GB/s   | ~2.6GB/s
    # AMD 3900XT     | ~20GB/s   | ~5GB/s
    #
    # i.e., ISA-L is consistently 3-5x faster than kernel sockets
    selected = crc32c_isal or crc32c_kern or crc32c_anycrc or None
    if not selected:
        raise NotImplementedError(
            'no crc32c implementation, install isal or anycrc')
    return selected


def _select_crc64nvme_impl():
    selected = crc64nvme_isal or crc64nvme_anycrc or None
    if not selected:
        raise NotImplementedError(
            'no crc64nvme implementation, install isal or anycrc')
    return selected


class CRCHasher(object):
    """
    Helper that works like a hashlib hasher, but with a CRC.
    """
    def __init__(self, name, crc_func, data=None, initial_value=0, width=32):
        """
        Initialize the CRCHasher.

        :param name: Name of the hasher
        :param crc_func: Function to compute the CRC.
        :param data: Data to update the hasher.
        :param initial_value: Initial CRC value.
        :param width: Width (in bits) of CRC values.
        """
        self.name = name
        self.crc_func = crc_func
        self.crc = initial_value
        if width not in (32, 64):
            raise ValueError("CRCHasher only supports 32- or 64-bit CRCs")
        self.width = width
        if data is not None:
            self.update(data)

    @property
    def digest_size(self):
        return self.width / 8

    @property
    def digest_fmt(self):
        return "!I" if self.width == 32 else "!Q"

    def update(self, data):
        """
        Update the CRC with new data.

        :param data: Data to update the CRC with.
        """
        self.crc = self.crc_func(data, self.crc)

    def digest(self):
        """
        Return the current CRC value as a big-endian integer of length
        ``width / 8`` bytes.

        :returns: Packed CRC value. (bytes)
        """
        return struct.pack(self.digest_fmt, self.crc)

    def hexdigest(self):
        """
        Return the hexadecimal representation of the current CRC value.

        :returns: Hexadecimal CRC value. (str)
        """
        hex = binascii.hexlify(self.digest()).decode("ascii")
        return hex

    def copy(self):
        """
        Copy the current state of this CRCHasher to a new one.

        :returns:
        """
        return CRCHasher(self.name,
                         self.crc_func,
                         initial_value=self.crc,
                         width=self.width)


def crc32(data=None, initial_value=0):
    return CRCHasher('crc32',
                     zlib.crc32,
                     data=data,
                     initial_value=initial_value)


def crc32c(data=None, initial_value=0):
    return CRCHasher('crc32c',
                     _select_crc32c_impl(),
                     data=data,
                     initial_value=initial_value)


def crc64nvme(data=None, initial_value=0):
    return CRCHasher('crc64nvme',
                     _select_crc64nvme_impl(),
                     data=data,
                     initial_value=initial_value,
                     width=64)


def log_selected_implementation(logger):
    try:
        impl = _select_crc32c_impl()
    except NotImplementedError:
        logger.warning(
            'No implementation found for CRC32C; '
            'install ISA-L or anycrc for support.')
    else:
        logger.info('Using %s implementation for CRC32C.' % impl.__name__)

    try:
        impl = _select_crc64nvme_impl()
    except NotImplementedError:
        logger.warning(
            'No implementation found for CRC64NVME; '
            'install ISA-L or anycrc for support.')
    else:
        logger.info('Using %s implementation for CRC64NVME.' % impl.__name__)



# Copyright (c) 2019 SwiftStack, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import logging
import os
import unittest
import uuid
import time

import boto3
from botocore.exceptions import ClientError
import urllib.parse

from swift.common.utils import config_true_value, readconf

from test import get_config

_CONFIG = None


# boto's loggign can get pretty noisy; require opt-in to see it all
if not config_true_value(os.environ.get('BOTO3_DEBUG')):
    logging.getLogger('boto3').setLevel(logging.INFO)
    logging.getLogger('botocore').setLevel(logging.INFO)


class ConfigError(Exception):
    '''Error test conf misconfigurations'''


def load_aws_config(conf_file):
    """
    Read user credentials from an AWS CLI style credentials file and translate
    to a swift test config. Currently only supports a single user.

    :param conf_file: path to AWS credentials file
    """
    conf = readconf(conf_file, 'default')
    global _CONFIG
    _CONFIG = {
        'endpoint': conf.get('endpoint', 'https://s3.amazonaws.com'),
        'region': conf.get('region', 'us-east-1'),
        'access_key1': conf.get('aws_access_key_id'),
        'secret_key1': conf.get('aws_secret_access_key'),
        'session_token1': conf.get('aws_session_token')
    }


aws_config_file = os.environ.get('SWIFT_TEST_AWS_CONFIG_FILE')
if aws_config_file:
    load_aws_config(aws_config_file)
    print('Loaded test config from %s' % aws_config_file)


def get_opt_or_error(option):
    global _CONFIG
    if _CONFIG is None:
        _CONFIG = get_config('s3api_test')

    value = _CONFIG.get(option)
    if not value:
        raise ConfigError('must supply [s3api_test]%s' % option)
    return value


def get_opt(option, default=None):
    try:
        return get_opt_or_error(option)
    except ConfigError:
        return default


def get_s3_client(user=1, signature_version='s3v4', addressing_style='path'):
    '''
    Get a boto3 client to talk to an S3 endpoint.

    :param user: user number to use. Should be one of:
        1 -- primary user
        2 -- secondary user
        3 -- unprivileged user
    :param signature_version: S3 signing method. Should be one of:
        s3 -- v2 signatures; produces Authorization headers like
              ``AWS access_key:signature``
        s3-query -- v2 pre-signed URLs; produces query strings like
                    ``?AWSAccessKeyId=access_key&Signature=signature``
        s3v4 -- v4 signatures; produces Authorization headers like
                ``AWS4-HMAC-SHA256
                Credential=access_key/date/region/s3/aws4_request,
                Signature=signature``
        s3v4-query -- v4 pre-signed URLs; produces query strings like
                      ``?X-Amz-Algorithm=AWS4-HMAC-SHA256&
                      X-Amz-Credential=access_key/date/region/s3/aws4_request&
                      X-Amz-Signature=signature``
    :param addressing_style: One of:
        path -- produces URLs like ``http(s)://host.domain/bucket/key``
        virtual -- produces URLs like ``http(s)://bucket.host.domain/key``
    '''
    endpoint = get_opt('endpoint', None)
    if endpoint:
        scheme = urllib.parse.urlsplit(endpoint).scheme
        if scheme not in ('http', 'https'):
            raise ConfigError('unexpected scheme in endpoint: %r; '
                              'expected http or https' % scheme)
    else:
        scheme = None
    region = get_opt('region', 'us-east-1')
    access_key = get_opt_or_error('access_key%d' % user)
    secret_key = get_opt_or_error('secret_key%d' % user)
    session_token = get_opt('session_token%d' % user)

    ca_cert = get_opt('ca_cert')
    if ca_cert is not None:
        try:
            # do a quick check now; it's more expensive to have boto check
            os.stat(ca_cert)
        except OSError as e:
            raise ConfigError(str(e))

    return boto3.client(
        's3',
        endpoint_url=endpoint,
        region_name=region,
        use_ssl=(scheme == 'https'),
        verify=ca_cert,
        config=boto3.session.Config(s3={
            'signature_version': signature_version,
            'addressing_style': addressing_style,
        }),
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key,
        aws_session_token=session_token
    )


def is_s3_acl_tests_enabled():
    explicit_opt = get_opt('s3_acl_tests_enabled', None)
    if explicit_opt is not None:
        return config_true_value(explicit_opt)
    else:
        legacy_opt = get_opt('s3_acl_tests_disabled', 'false')
        return not config_true_value(legacy_opt)


def skip_if_s3_acl_tests_disabled(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not is_s3_acl_tests_enabled():
            raise unittest.SkipTest('s3_acl_tests_enabled is false')
        return func(*args, **kwargs)
    return wrapper


def etag_from_resp(response):
    return response['ETag']


def code_from_error(error):
    return error.response['Error']['Code']


def status_from_error(error):
    return error.response['ResponseMetadata']['HTTPStatusCode']


TEST_PREFIX = 's3api-test-'


class BaseS3Mixin(object):
    # Default to v4 signatures (as aws-cli does), but subclasses can override
    signature_version = 's3v4'

    @classmethod
    def get_s3_client(cls, user):
        return get_s3_client(user, cls.signature_version)

    @classmethod
    def _remove_all_object_versions_from_bucket(cls, client, bucket_name):
        resp = client.list_object_versions(Bucket=bucket_name)
        objs_to_delete = (resp.get('Versions', []) +
                          resp.get('DeleteMarkers', []))
        while objs_to_delete:
            multi_delete_body = {
                'Objects': [
                    {'Key': obj['Key'], 'VersionId': obj['VersionId']}
                    for obj in objs_to_delete
                ],
                'Quiet': False,
            }
            del_resp = client.delete_objects(Bucket=bucket_name,
                                             Delete=multi_delete_body)
            if any(del_resp.get('Errors', [])):
                raise Exception('Unable to delete %r' % del_resp['Errors'])
            if not resp['IsTruncated']:
                break
            key_marker = resp['NextKeyMarker']
            version_id_marker = resp['NextVersionIdMarker']
            resp = client.list_object_versions(
                Bucket=bucket_name, KeyMarker=key_marker,
                VersionIdMarker=version_id_marker)
            objs_to_delete = (resp.get('Versions', []) +
                              resp.get('DeleteMarkers', []))

    @classmethod
    def clear_bucket(cls, client, bucket_name):
        timeout = time.time() + 10
        backoff = 0.1
        cls._remove_all_object_versions_from_bucket(client, bucket_name)
        try:
            client.delete_bucket(Bucket=bucket_name)
        except ClientError as e:
            if 'NoSuchBucket' in str(e):
                return
            if 'BucketNotEmpty' not in str(e):
                raise
            # Something's gone sideways. Try harder
            client.put_bucket_versioning(
                Bucket=bucket_name,
                VersioningConfiguration={'Status': 'Suspended'})
            while True:
                cls._remove_all_object_versions_from_bucket(
                    client, bucket_name)
                # also try some version-unaware operations...
                for key in client.list_objects(Bucket=bucket_name).get(
                        'Contents', []):
                    client.delete_object(Bucket=bucket_name, Key=key['Key'])

                # *then* try again
                try:
                    client.delete_bucket(Bucket=bucket_name)
                except ClientError as e:
                    if 'NoSuchBucket' in str(e):
                        return
                    if 'BucketNotEmpty' not in str(e):
                        raise
                    if time.time() > timeout:
                        raise Exception('Timeout clearing %r' % bucket_name)
                    time.sleep(backoff)
                    backoff *= 2
                else:
                    break

    @classmethod
    def create_name(cls, slug):
        return '%s%s-%s' % (TEST_PREFIX, slug, uuid.uuid4().hex)

    @classmethod
    def clear_account(cls, client):
        for bucket in client.list_buckets()['Buckets']:
            if not bucket['Name'].startswith(TEST_PREFIX):
                # these tests run against real s3 accounts
                continue
            cls.clear_bucket(client, bucket['Name'])


class BaseS3TestCase(BaseS3Mixin, unittest.TestCase):
    def tearDown(self):
        client = self.get_s3_client(1)
        self.clear_account(client)
        try:
            client = self.get_s3_client(2)
        except ConfigError:
            pass
        else:
            self.clear_account(client)

    def check_owner(self, owner):
        # as of July 2025, S3 may or may not send DisplayName
        # https://docs.aws.amazon.com/AmazonS3/latest/API/API_Owner.html
        owner.pop('DisplayName', None)
        self.assertEqual(['ID'], list(owner.keys()))


class BaseS3TestCaseWithBucket(BaseS3Mixin, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.bucket_name = cls.create_name('bucket')
        client = cls.get_s3_client(1)
        client.create_bucket(Bucket=cls.bucket_name)

    @classmethod
    def tearDownClass(cls):
        client = cls.get_s3_client(1)
        cls.clear_account(client)
        try:
            client = cls.get_s3_client(2)
        except ConfigError:
            pass
        else:
            cls.clear_account(client)



# Copyright (c) 2010-2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import time
import functools

from swift import __version__ as swift_version
from swift.common.utils import public, config_true_value, \
    LOG_LINE_DEFAULT_FORMAT
from swift.common.http import is_server_error
from swift.common.swob import Response, HTTPException


def labeled_timing_stats(metric, **dec_kwargs):
    """
    Returns a decorator that emits labeled metrics timing events or errors
    for public methods in swift's wsgi server controllers, based on response
    code.

    The controller methods are not allowed to override the following labels:
    'method', 'status'.
    """
    def decorating_func(func):

        @functools.wraps(func)
        def _timing_stats(ctrl, req, *args, **kwargs):
            labels = {}
            start_time = time.time()
            req_method = req.method
            try:
                resp = func(
                    ctrl, req, *args, timing_stats_labels=labels, **kwargs)
            except HTTPException as e:
                resp = e
            labels['method'] = req_method
            labels['status'] = resp.status_int

            ctrl.statsd.timing_since(metric, start_time, labels=labels,
                                     **dec_kwargs)
            return resp

        return _timing_stats
    return decorating_func


def timing_stats(**dec_kwargs):
    """
    Returns a decorator that logs timing events or errors for public methods in
    swift's wsgi server controllers, based on response code.
    """
    def decorating_func(func):
        method = func.__name__

        @functools.wraps(func)
        def _timing_stats(ctrl, *args, **kwargs):
            start_time = time.time()
            try:
                resp = func(ctrl, *args, **kwargs)
            except HTTPException as e:
                resp = e
            # .timing is for successful responses *or* error codes that are
            # not Swift's fault. For example, 500 is definitely the server's
            # fault, but 412 is an error code (4xx are all errors) that is
            # due to a header the client sent.
            #
            # .errors.timing is for failures that *are* Swift's fault.
            # Examples include 507 for an unmounted drive or 500 for an
            # unhandled exception.
            if not is_server_error(resp.status_int):
                ctrl.logger.timing_since(method + '.timing',
                                         start_time, **dec_kwargs)
            else:
                ctrl.logger.timing_since(method + '.errors.timing',
                                         start_time, **dec_kwargs)
            return resp

        return _timing_stats
    return decorating_func


class BaseStorageServer(object):
    """
    Implements common OPTIONS method for object, account, container servers.
    """

    def __init__(self, conf, **kwargs):
        self._allowed_methods = None
        self.replication_server = config_true_value(
            conf.get('replication_server', 'true'))
        self.log_format = conf.get('log_format', LOG_LINE_DEFAULT_FORMAT)
        self.anonymization_method = conf.get('log_anonymization_method', 'md5')
        self.anonymization_salt = conf.get('log_anonymization_salt', '')

    @property
    def server_type(self):
        raise NotImplementedError(
            'Storage nodes have not implemented the Server type.')

    @property
    def allowed_methods(self):
        if self._allowed_methods is None:
            self._allowed_methods = []
            all_methods = inspect.getmembers(self, predicate=callable)
            for name, m in all_methods:
                if not getattr(m, 'publicly_accessible', False):
                    continue
                if getattr(m, 'replication', False) and \
                        not self.replication_server:
                    continue
                self._allowed_methods.append(name)
            self._allowed_methods.sort()
        return self._allowed_methods

    @public
    @timing_stats()
    def OPTIONS(self, req):
        """
        Base handler for OPTIONS requests

        :param req: swob.Request object
        :returns: swob.Response object
        """
        # Prepare the default response
        headers = {'Allow': ', '.join(self.allowed_methods),
                   'Server': '%s/%s' % (self.server_type, swift_version)}
        resp = Response(status=200, request=req, headers=headers)

        return resp



# Copyright (c) 2019 SwiftStack, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from collections import defaultdict

from botocore.exceptions import ClientError
import io

from swift.common.header_key_dict import HeaderKeyDict
from swift.common.utils import md5
from test.s3api import BaseS3TestCase


def retry(f, timeout=10):
    timelimit = time.time() + timeout
    while True:
        try:
            f()
        except (ClientError, AssertionError):
            if time.time() > timelimit:
                raise
            continue
        else:
            break


class TestObjectVersioning(BaseS3TestCase):

    maxDiff = None

    def _sanitize_obj_listing(self, obj):
        # there's some object listing parameters that are not deterministic
        obj.pop('LastModified')
        obj.pop('Owner', None)
        # there's some object listing parameters that Swift doesn't return,
        obj.pop('ChecksumAlgorithm', None)
        obj.pop('ChecksumType', None)

    def setUp(self):
        self.client = self.get_s3_client(1)
        self.bucket_name = self.create_name('versioning')
        resp = self.client.create_bucket(Bucket=self.bucket_name)
        self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])

        def enable_versioning():
            resp = self.client.put_bucket_versioning(
                Bucket=self.bucket_name,
                VersioningConfiguration={'Status': 'Enabled'})
            self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])
        retry(enable_versioning)

    def tearDown(self):
        resp = self.client.put_bucket_versioning(
            Bucket=self.bucket_name,
            VersioningConfiguration={'Status': 'Suspended'})
        self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])
        self.clear_bucket(self.client, self.bucket_name)
        super(TestObjectVersioning, self).tearDown()

    def test_setup(self):
        bucket_name = self.create_name('new-bucket')
        resp = self.client.create_bucket(Bucket=bucket_name)
        self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])
        expected_location = '/%s' % bucket_name
        self.assertEqual(expected_location, resp['Location'])
        headers = HeaderKeyDict(resp['ResponseMetadata']['HTTPHeaders'])
        self.assertEqual('0', headers['content-length'])
        self.assertEqual(expected_location, headers['location'])

        # get versioning
        resp = self.client.get_bucket_versioning(Bucket=bucket_name)
        self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])
        self.assertNotIn('Status', resp)

        # put versioning
        versioning_config = {
            'Status': 'Enabled',
        }
        resp = self.client.put_bucket_versioning(
            Bucket=bucket_name,
            VersioningConfiguration=versioning_config)
        self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])

        # ... now it's enabled
        def check_status():
            resp = self.client.get_bucket_versioning(Bucket=bucket_name)
            self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])
            try:
                self.assertEqual('Enabled', resp['Status'])
            except KeyError:
                self.fail('Status was not in %r' % resp)
        retry(check_status)

        # send over some bogus junk
        versioning_config['Status'] = 'Disabled'
        with self.assertRaises(ClientError) as ctx:
            self.client.put_bucket_versioning(
                Bucket=bucket_name,
                VersioningConfiguration=versioning_config)
        expected_err = 'An error occurred (MalformedXML) when calling the ' \
            'PutBucketVersioning operation: The XML you provided was ' \
            'not well-formed or did not validate against our published schema'
        self.assertEqual(expected_err, str(ctx.exception))

        # disable it
        versioning_config['Status'] = 'Suspended'
        resp = self.client.put_bucket_versioning(
            Bucket=bucket_name,
            VersioningConfiguration=versioning_config)
        self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])

        # ... now it's disabled again
        def check_status():
            resp = self.client.get_bucket_versioning(Bucket=bucket_name)
            self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])
            self.assertEqual('Suspended', resp['Status'])
        retry(check_status)

    def test_upload_fileobj_versioned(self):
        obj_data = self.create_name('some-data').encode('ascii')
        obj_etag = md5(obj_data, usedforsecurity=False).hexdigest()
        obj_name = self.create_name('versioned-obj')
        self.client.upload_fileobj(io.BytesIO(obj_data),
                                   self.bucket_name, obj_name)

        # object is in the listing
        resp = self.client.list_objects_v2(Bucket=self.bucket_name)
        objs = resp.get('Contents', [])
        for obj in objs:
            self._sanitize_obj_listing(obj)
        self.assertEqual([{
            'ETag': '"%s"' % obj_etag,
            'Key': obj_name,
            'Size': len(obj_data),
            'StorageClass': 'STANDARD',
        }], objs)

        # object version listing
        resp = self.client.list_object_versions(Bucket=self.bucket_name)
        objs = resp.get('Versions', [])
        for obj in objs:
            self._sanitize_obj_listing(obj)
            obj.pop('VersionId')
        self.assertEqual([{
            'ETag': '"%s"' % obj_etag,
            'IsLatest': True,
            'Key': obj_name,
            'Size': len(obj_data),
            'StorageClass': 'STANDARD',
        }], objs)

        # overwrite the object
        new_obj_data = self.create_name('some-new-data').encode('ascii')
        new_obj_etag = md5(new_obj_data, usedforsecurity=False).hexdigest()
        self.client.upload_fileobj(io.BytesIO(new_obj_data),
                                   self.bucket_name, obj_name)

        # new object is in the listing
        resp = self.client.list_objects_v2(Bucket=self.bucket_name)
        objs = resp.get('Contents', [])
        for obj in objs:
            self._sanitize_obj_listing(obj)
        self.assertEqual([{
            'ETag': '"%s"' % new_obj_etag,
            'Key': obj_name,
            'Size': len(new_obj_data),
            'StorageClass': 'STANDARD',
        }], objs)

        # both object versions in the versions listing
        resp = self.client.list_object_versions(Bucket=self.bucket_name)
        objs = resp.get('Versions', [])
        for obj in objs:
            self._sanitize_obj_listing(obj)
            obj.pop('VersionId')
        self.assertEqual([{
            'ETag': '"%s"' % new_obj_etag,
            'IsLatest': True,
            'Key': obj_name,
            'Size': len(new_obj_data),
            'StorageClass': 'STANDARD',
        }, {
            'ETag': '"%s"' % obj_etag,
            'IsLatest': False,
            'Key': obj_name,
            'Size': len(obj_data),
            'StorageClass': 'STANDARD',
        }], objs)

    def test_delete_versioned_objects(self):
        etags = []
        obj_name = self.create_name('versioned-obj')
        for i in range(3):
            obj_data = self.create_name('some-data-%s' % i).encode('ascii')
            etags.insert(0, md5(obj_data, usedforsecurity=False).hexdigest())
            self.client.upload_fileobj(io.BytesIO(obj_data),
                                       self.bucket_name, obj_name)

        # only one object appears in the listing
        resp = self.client.list_objects_v2(Bucket=self.bucket_name)
        objs = resp.get('Contents', [])
        for obj in objs:
            self._sanitize_obj_listing(obj)
        self.assertEqual([{
            'ETag': '"%s"' % etags[0],
            'Key': obj_name,
            'Size': len(obj_data),
            'StorageClass': 'STANDARD',
        }], objs)

        # but everything is layed out in the object versions listing
        resp = self.client.list_object_versions(Bucket=self.bucket_name)
        objs = resp.get('Versions', [])
        versions = []
        for obj in objs:
            self._sanitize_obj_listing(obj)
            versions.append(obj.pop('VersionId'))
        self.assertEqual([{
            'ETag': '"%s"' % etags[0],
            'IsLatest': True,
            'Key': obj_name,
            'Size': len(obj_data),
            'StorageClass': 'STANDARD',
        }, {
            'ETag': '"%s"' % etags[1],
            'IsLatest': False,
            'Key': obj_name,
            'Size': len(obj_data),
            'StorageClass': 'STANDARD',
        }, {
            'ETag': '"%s"' % etags[2],
            'IsLatest': False,
            'Key': obj_name,
            'Size': len(obj_data),
            'StorageClass': 'STANDARD',
        }], objs)

        # we can delete a specific version
        resp = self.client.delete_object(Bucket=self.bucket_name,
                                         Key=obj_name,
                                         VersionId=versions[1])

        # and that just pulls it out of the versions listing
        resp = self.client.list_object_versions(Bucket=self.bucket_name)
        objs = resp.get('Versions', [])
        for obj in objs:
            self._sanitize_obj_listing(obj)
            obj.pop('VersionId')
        self.assertEqual([{
            'ETag': '"%s"' % etags[0],
            'IsLatest': True,
            'Key': obj_name,
            'Size': len(obj_data),
            'StorageClass': 'STANDARD',
        }, {
            'ETag': '"%s"' % etags[2],
            'IsLatest': False,
            'Key': obj_name,
            'Size': len(obj_data),
            'StorageClass': 'STANDARD',
        }], objs)

        # ... but the current listing is unaffected
        resp = self.client.list_objects_v2(Bucket=self.bucket_name)
        objs = resp.get('Contents', [])
        for obj in objs:
            self._sanitize_obj_listing(obj)
        self.assertEqual([{
            'ETag': '"%s"' % etags[0],
            'Key': obj_name,
            'Size': len(obj_data),
            'StorageClass': 'STANDARD',
        }], objs)

        # OTOH, if you delete specifically the latest version
        # we can delete a specific version
        resp = self.client.delete_object(Bucket=self.bucket_name,
                                         Key=obj_name,
                                         VersionId=versions[0])

        # the versions listing has a new IsLatest
        resp = self.client.list_object_versions(Bucket=self.bucket_name)
        objs = resp.get('Versions', [])
        for obj in objs:
            self._sanitize_obj_listing(obj)
            obj.pop('VersionId')
        self.assertEqual([{
            'ETag': '"%s"' % etags[2],
            'IsLatest': True,
            'Key': obj_name,
            'Size': len(obj_data),
            'StorageClass': 'STANDARD',
        }], objs)

        # and the stack pops
        resp = self.client.list_objects_v2(Bucket=self.bucket_name)
        objs = resp.get('Contents', [])
        for obj in objs:
            self._sanitize_obj_listing(obj)
        self.assertEqual([{
            'ETag': '"%s"' % etags[2],
            'Key': obj_name,
            'Size': len(obj_data),
            'StorageClass': 'STANDARD',
        }], objs)

    def test_delete_versioned_deletes(self):
        etags = []
        obj_name = self.create_name('versioned-obj')
        for i in range(3):
            obj_data = self.create_name('some-data-%s' % i).encode('ascii')
            etags.insert(0, md5(obj_data, usedforsecurity=False).hexdigest())
            self.client.upload_fileobj(io.BytesIO(obj_data),
                                       self.bucket_name, obj_name)
            # and make a delete marker
            self.client.delete_object(Bucket=self.bucket_name, Key=obj_name)

        # current listing is empty
        resp = self.client.list_objects_v2(Bucket=self.bucket_name)
        objs = resp.get('Contents', [])
        self.assertEqual([], objs)

        # but everything is in layed out in the versions listing
        resp = self.client.list_object_versions(Bucket=self.bucket_name)
        objs = resp.get('Versions', [])
        versions = []
        for obj in objs:
            self._sanitize_obj_listing(obj)
            versions.append(obj.pop('VersionId'))
        self.assertEqual([{
            'ETag': '"%s"' % etag,
            'IsLatest': False,
            'Key': obj_name,
            'Size': len(obj_data),
            'StorageClass': 'STANDARD',
        } for etag in etags], objs)
        # ... plus the delete markers
        delete_markers = resp.get('DeleteMarkers', [])
        marker_versions = []
        for marker in delete_markers:
            self._sanitize_obj_listing(marker)
            marker_versions.append(marker.pop('VersionId'))
        self.assertEqual([{
            'Key': obj_name,
            'IsLatest': is_latest,
        } for is_latest in (True, False, False)], delete_markers)

        # delete an old delete markers
        resp = self.client.delete_object(Bucket=self.bucket_name,
                                         Key=obj_name,
                                         VersionId=marker_versions[2])

        # since IsLatest is still marker we'll raise NoSuchKey
        with self.assertRaises(ClientError) as caught:
            resp = self.client.get_object(Bucket=self.bucket_name,
                                          Key=obj_name)
        expected_err = 'An error occurred (NoSuchKey) when calling the ' \
            'GetObject operation: The specified key does not exist.'
        self.assertEqual(expected_err, str(caught.exception))

        # now delete the delete marker (IsLatest)
        resp = self.client.delete_object(Bucket=self.bucket_name,
                                         Key=obj_name,
                                         VersionId=marker_versions[0])

        # most recent version is now latest
        resp = self.client.get_object(Bucket=self.bucket_name,
                                      Key=obj_name)
        self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])
        self.assertEqual('"%s"' % etags[0], resp['ETag'])

        # now delete the IsLatest object version
        resp = self.client.delete_object(Bucket=self.bucket_name,
                                         Key=obj_name,
                                         VersionId=versions[0])

        # and object is deleted again
        with self.assertRaises(ClientError) as caught:
            resp = self.client.get_object(Bucket=self.bucket_name,
                                          Key=obj_name)
        expected_err = 'An error occurred (NoSuchKey) when calling the ' \
            'GetObject operation: The specified key does not exist.'
        self.assertEqual(expected_err, str(caught.exception))

        # delete marker IsLatest
        resp = self.client.list_object_versions(Bucket=self.bucket_name)
        delete_markers = resp.get('DeleteMarkers', [])
        for marker in delete_markers:
            self._sanitize_obj_listing(marker)
        self.assertEqual([{
            'Key': obj_name,
            'IsLatest': True,
            'VersionId': marker_versions[1],
        }], delete_markers)

    def test_multipart_upload(self):
        obj_name = self.create_name('versioned-obj')
        obj_data = b'data'

        mu = self.client.create_multipart_upload(
            Bucket=self.bucket_name,
            Key=obj_name)
        part_md5 = self.client.upload_part(
            Bucket=self.bucket_name,
            Key=obj_name,
            UploadId=mu['UploadId'],
            PartNumber=1,
            Body=obj_data)['ETag']
        complete_response = self.client.complete_multipart_upload(
            Bucket=self.bucket_name,
            Key=obj_name,
            UploadId=mu['UploadId'],
            MultipartUpload={'Parts': [
                {'PartNumber': 1, 'ETag': part_md5},
            ]})
        obj_etag = complete_response['ETag']

        delete_response = self.client.delete_object(
            Bucket=self.bucket_name,
            Key=obj_name)
        marker_version_id = delete_response['VersionId']

        resp = self.client.list_object_versions(Bucket=self.bucket_name)
        objs = resp.get('Versions', [])
        versions = []
        for obj in objs:
            self._sanitize_obj_listing(obj)
            versions.append(obj.pop('VersionId'))
        self.assertEqual([{
            'ETag': obj_etag,
            'IsLatest': False,
            'Key': obj_name,
            'Size': len(obj_data),
            'StorageClass': 'STANDARD',
        }], objs)

        markers = resp.get('DeleteMarkers', [])
        for marker in markers:
            marker.pop('LastModified')
            marker.pop('Owner')
        self.assertEqual([{
            'IsLatest': True,
            'Key': obj_name,
            'VersionId': marker_version_id,
        }], markers)

        # Can still get the old version
        resp = self.client.get_object(
            Bucket=self.bucket_name,
            Key=obj_name,
            VersionId=versions[0])
        self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])
        self.assertEqual(obj_etag, resp['ETag'])

        delete_response = self.client.delete_object(
            Bucket=self.bucket_name,
            Key=obj_name,
            VersionId=versions[0])

        resp = self.client.list_object_versions(Bucket=self.bucket_name)
        self.assertEqual([], resp.get('Versions', []))

        markers = resp.get('DeleteMarkers', [])
        for marker in markers:
            self._sanitize_obj_listing(marker)
        self.assertEqual([{
            'IsLatest': True,
            'Key': obj_name,
            'VersionId': marker_version_id,
        }], markers)

    def test_get_versioned_object(self):
        etags = []
        obj_name = self.create_name('versioned-obj')
        for i in range(3):
            obj_data = self.create_name('some-data-%s' % i).encode('ascii')
            # TODO: pull etag from response instead
            etags.insert(0, md5(obj_data, usedforsecurity=False).hexdigest())
            self.client.upload_fileobj(
                io.BytesIO(obj_data), self.bucket_name, obj_name)

        resp = self.client.list_object_versions(Bucket=self.bucket_name)
        objs = resp.get('Versions', [])
        versions = []
        for obj in objs:
            self._sanitize_obj_listing(obj)
            versions.append(obj.pop('VersionId'))
        self.assertEqual([{
            'ETag': '"%s"' % etags[0],
            'IsLatest': True,
            'Key': obj_name,
            'Size': len(obj_data),
            'StorageClass': 'STANDARD',
        }, {
            'ETag': '"%s"' % etags[1],
            'IsLatest': False,
            'Key': obj_name,
            'Size': len(obj_data),
            'StorageClass': 'STANDARD',
        }, {
            'ETag': '"%s"' % etags[2],
            'IsLatest': False,
            'Key': obj_name,
            'Size': len(obj_data),
            'StorageClass': 'STANDARD',
        }], objs)

        # un-versioned get_object returns IsLatest
        resp = self.client.get_object(Bucket=self.bucket_name, Key=obj_name)
        self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])
        self.assertEqual('"%s"' % etags[0], resp['ETag'])

        # but you can get any object by version
        for i, version in enumerate(versions):
            resp = self.client.get_object(
                Bucket=self.bucket_name, Key=obj_name, VersionId=version)
            self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])
            self.assertEqual('"%s"' % etags[i], resp['ETag'])

        # and head_object works about the same
        resp = self.client.head_object(Bucket=self.bucket_name, Key=obj_name)
        self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])
        self.assertEqual('"%s"' % etags[0], resp['ETag'])
        self.assertEqual(versions[0], resp['VersionId'])
        for version, etag in zip(versions, etags):
            resp = self.client.head_object(
                Bucket=self.bucket_name, Key=obj_name, VersionId=version)
            self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])
            self.assertEqual(version, resp['VersionId'])
            self.assertEqual('"%s"' % etag, resp['ETag'])

    def test_get_versioned_object_invalid_params(self):
        with self.assertRaises(ClientError) as ctx:
            self.client.list_object_versions(Bucket=self.bucket_name,
                                             KeyMarker='',
                                             VersionIdMarker='bogus')
        expected_err = 'An error occurred (InvalidArgument) when calling ' \
            'the ListObjectVersions operation: Invalid version id specified'
        self.assertEqual(expected_err, str(ctx.exception))

        with self.assertRaises(ClientError) as ctx:
            self.client.list_object_versions(
                Bucket=self.bucket_name,
                VersionIdMarker='a' * 32)
        expected_err = 'An error occurred (InvalidArgument) when calling ' \
            'the ListObjectVersions operation: A version-id marker cannot ' \
            'be specified without a key marker.'
        self.assertEqual(expected_err, str(ctx.exception))

    def test_get_versioned_object_key_marker(self):
        obj00_name = self.create_name('00-versioned-obj')
        obj01_name = self.create_name('01-versioned-obj')
        names = [obj00_name] * 3 + [obj01_name] * 3
        latest = [True, False, False, True, False, False]
        etags = []
        for i in range(3):
            obj_data = self.create_name('some-data-%s' % i).encode('ascii')
            etags.insert(0, '"%s"' % md5(
                obj_data, usedforsecurity=False).hexdigest())
            self.client.upload_fileobj(
                io.BytesIO(obj_data), self.bucket_name, obj01_name)
        for i in range(3):
            obj_data = self.create_name('some-data-%s' % i).encode('ascii')
            etags.insert(0, '"%s"' % md5(
                obj_data, usedforsecurity=False).hexdigest())
            self.client.upload_fileobj(
                io.BytesIO(obj_data), self.bucket_name, obj00_name)
        resp = self.client.list_object_versions(Bucket=self.bucket_name)
        versions = []
        objs = []
        for o in resp.get('Versions', []):
            versions.append(o['VersionId'])
            objs.append({
                'Key': o['Key'],
                'VersionId': o['VersionId'],
                'IsLatest': o['IsLatest'],
                'ETag': o['ETag'],
            })
        expected = [{
            'Key': name,
            'VersionId': version,
            'IsLatest': is_latest,
            'ETag': etag,
        } for name, etag, version, is_latest in zip(
            names, etags, versions, latest)]
        self.assertEqual(expected, objs)

        # on s3 this makes expected[0]['IsLatest'] magicaly change to False?
        # resp = self.client.list_object_versions(Bucket=self.bucket_name,
        #                                         KeyMarker='',
        #                                         VersionIdMarker=versions[0])
        # objs = [{
        #     'Key': o['Key'],
        #     'VersionId': o['VersionId'],
        #     'IsLatest': o['IsLatest'],
        #     'ETag': o['ETag'],
        # } for o in resp.get('Versions', [])]
        # self.assertEqual(expected, objs)

        # KeyMarker skips past that key
        resp = self.client.list_object_versions(Bucket=self.bucket_name,
                                                KeyMarker=obj00_name)
        objs = [{
            'Key': o['Key'],
            'VersionId': o['VersionId'],
            'IsLatest': o['IsLatest'],
            'ETag': o['ETag'],
        } for o in resp.get('Versions', [])]
        self.assertEqual(expected[3:], objs)

        # KeyMarker with VersionIdMarker skips past that version
        resp = self.client.list_object_versions(Bucket=self.bucket_name,
                                                KeyMarker=obj00_name,
                                                VersionIdMarker=versions[0])
        objs = [{
            'Key': o['Key'],
            'VersionId': o['VersionId'],
            'IsLatest': o['IsLatest'],
            'ETag': o['ETag'],
        } for o in resp.get('Versions', [])]
        self.assertEqual(expected[1:], objs)

        # KeyMarker with bogus version skips past that key
        resp = self.client.list_object_versions(
            Bucket=self.bucket_name,
            KeyMarker=obj00_name,
            VersionIdMarker=versions[4])
        objs = [{
            'Key': o['Key'],
            'VersionId': o['VersionId'],
            'IsLatest': o['IsLatest'],
            'ETag': o['ETag'],
        } for o in resp.get('Versions', [])]
        self.assertEqual(expected[3:], objs)

    def test_list_objects(self):
        etags = defaultdict(list)
        for i in range(3):
            obj_name = self.create_name('versioned-obj')
            for i in range(3):
                obj_data = self.create_name('some-data-%s' % i).encode('ascii')
                etags[obj_name].insert(0, md5(
                    obj_data, usedforsecurity=False).hexdigest())
                self.client.upload_fileobj(
                    io.BytesIO(obj_data), self.bucket_name, obj_name)

        # both unversioned list_objects responses are similar
        expected = []
        for name, obj_etags in sorted(etags.items()):
            expected.append({
                'ETag': '"%s"' % obj_etags[0],
                'Key': name,
                'Size': len(obj_data),
                'StorageClass': 'STANDARD',
            })
        resp = self.client.list_objects(Bucket=self.bucket_name)
        objs = resp.get('Contents', [])
        for obj in objs:
            owner = obj.pop('Owner')
            # one difference seems to be the Owner key
            self.check_owner(owner)
            self._sanitize_obj_listing(obj)
        self.assertEqual(expected, objs)
        resp = self.client.list_objects_v2(Bucket=self.bucket_name)
        objs = resp.get('Contents', [])
        for obj in objs:
            self._sanitize_obj_listing(obj)
        self.assertEqual(expected, objs)

        # versioned listings has something for everyone
        expected = []
        for name, obj_etags in sorted(etags.items()):
            is_latest = True
            for etag in obj_etags:
                expected.append({
                    'ETag': '"%s"' % etag,
                    'IsLatest': is_latest,
                    'Key': name,
                    'Size': len(obj_data),
                    'StorageClass': 'STANDARD',
                })
                is_latest = False

        resp = self.client.list_object_versions(Bucket=self.bucket_name)
        objs = resp.get('Versions', [])
        versions = []
        for obj in objs:
            self._sanitize_obj_listing(obj)
            versions.append(obj.pop('VersionId'))
        self.assertEqual(expected, objs)

    def test_copy_object(self):
        etags = []
        obj_name = self.create_name('versioned-obj')
        for i in range(3):
            obj_data = self.create_name('some-data-%s' % i).encode('ascii')
            etags.insert(0, md5(
                obj_data, usedforsecurity=False).hexdigest())
            self.client.upload_fileobj(
                io.BytesIO(obj_data), self.bucket_name, obj_name)

        resp = self.client.list_object_versions(Bucket=self.bucket_name)
        objs = resp.get('Versions', [])
        versions = []
        for obj in objs:
            versions.append(obj.pop('VersionId'))

        # CopySource can just be Bucket/Key string
        first_target = self.create_name('target-obj1')
        copy_resp = self.client.copy_object(
            Bucket=self.bucket_name, Key=first_target,
            CopySource='%s/%s' % (self.bucket_name, obj_name))
        self.assertEqual(versions[0], copy_resp['CopySourceVersionId'])

        # and you'll just get the most recent version
        resp = self.client.head_object(Bucket=self.bucket_name,
                                       Key=first_target)
        self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])
        self.assertEqual('"%s"' % etags[0], resp['ETag'])

        # or you can be more explicit
        explicit_target = self.create_name('target-%s' % versions[0])
        copy_source = {'Bucket': self.bucket_name, 'Key': obj_name,
                       'VersionId': versions[0]}
        copy_resp = self.client.copy_object(
            Bucket=self.bucket_name, Key=explicit_target,
            CopySource=copy_source)
        self.assertEqual(versions[0], copy_resp['CopySourceVersionId'])
        # and you still get the same thing
        resp = self.client.head_object(Bucket=self.bucket_name,
                                       Key=explicit_target)
        self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])
        self.assertEqual('"%s"' % etags[0], resp['ETag'])

        # but you can also copy from a specific version
        version_target = self.create_name('target-%s' % versions[2])
        copy_source['VersionId'] = versions[2]
        copy_resp = self.client.copy_object(
            Bucket=self.bucket_name, Key=version_target,
            CopySource=copy_source)
        self.assertEqual(versions[2], copy_resp['CopySourceVersionId'])
        resp = self.client.head_object(Bucket=self.bucket_name,
                                       Key=version_target)
        self.assertEqual(200, resp['ResponseMetadata']['HTTPStatusCode'])
        self.assertEqual('"%s"' % etags[2], resp['ETag'])



# Copyright (c) 2010-2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Statsd Client """

import time
import warnings
import re
from contextlib import closing
from random import random

from eventlet.green import socket

from swift.common.utils.config import config_true_value


STATSD_CONF_USER_LABEL_PREFIX = 'statsd_user_label_'
STATSD_USER_LABEL_NAMESPACE = 'user_'
USER_LABEL_PATTERN = re.compile(r"[^0-9a-zA-Z_]")
USER_VALUE_PATTERN = re.compile(r"[^0-9a-zA-Z_.]")


def _build_line_parts(metric, value, metric_type, sample_rate):
    line = '%s:%s|%s' % (metric, value, metric_type)
    if sample_rate < 1:
        line += '|@%s' % (sample_rate,)
    return line


def librato(metric, value, metric_type, sample_rate, labels):
    # https://www.librato.com/docs/kb/collect/collection_agents/stastd/#stat-level-tags
    if labels:
        metric += '#' + ','.join('%s=%s' % (k, v) for k, v in labels)
    line = _build_line_parts(metric, value, metric_type, sample_rate)
    return line


def influxdb(metric, value, metric_type, sample_rate, labels):
    # https://www.influxdata.com/blog/getting-started-with-sending-statsd-metrics-to-telegraf-influxdb/#introducing-influx-statsd
    if labels:
        metric += ''.join(',%s=%s' % (k, v) for k, v in labels)
    line = _build_line_parts(metric, value, metric_type, sample_rate)
    return line


def graphite(metric, value, metric_type, sample_rate, labels):
    # https://graphite.readthedocs.io/en/latest/tags.html#carbon
    if labels:
        metric += ''.join(';%s=%s' % (k, v) for k, v in labels)
    line = _build_line_parts(metric, value, metric_type, sample_rate)
    return line


def dogstatsd(metric, value, metric_type, sample_rate, labels):
    # https://docs.datadoghq.com/developers/dogstatsd/datagram_shell/?tab=metrics
    line = _build_line_parts(metric, value, metric_type, sample_rate)
    if labels:
        line += '|#' + ','.join('%s:%s' % (k, v) for k, v in labels)
    return line


LABEL_MODES = {
    'disabled': None,
    'librato': librato,
    'influxdb': influxdb,
    'graphite': graphite,
    'dogstatsd': dogstatsd,
}


def _get_labeled_statsd_formatter(label_mode):
    """
    Returns a label formatting function for the given ``label_mode``.

    :param label_mode: A label mode.
    :raises ValueError: if ``label_mode`` is not supported by ``LabelFormats``.
    :returns: a label formatting function.
    """
    try:
        return LABEL_MODES[label_mode]
    except KeyError:
        label_modes = LABEL_MODES.keys()
        raise ValueError(
            'unknown statsd_label_mode %r; '
            'expected one of %r' % (label_mode, label_modes))


def get_statsd_client(conf=None, tail_prefix='', logger=None):
    """
    Get an instance of StatsdClient using config settings.

    **config and defaults**::

        log_statsd_host = (disabled)
        log_statsd_port = 8125
        log_statsd_default_sample_rate = 1.0
        log_statsd_sample_rate_factor = 1.0
        log_statsd_metric_prefix = (empty-string)
        statsd_emit_legacy = true

    :param conf: Configuration dict to read settings from
    :param tail_prefix: tail prefix to pass to statsd client
    :param logger: stdlib logger instance used by statsd client for logging
    :return: an instance of ``StatsdClient``
    """
    conf = conf or {}

    host = conf.get('log_statsd_host')
    port = int(conf.get('log_statsd_port', 8125))
    base_prefix = conf.get('log_statsd_metric_prefix', '')
    default_sample_rate = float(
        conf.get('log_statsd_default_sample_rate', 1))
    sample_rate_factor = float(
        conf.get('log_statsd_sample_rate_factor', 1))

    emit_legacy = config_true_value(conf.get(
        'statsd_emit_legacy', 'true'))

    return StatsdClient(
        host, port,
        base_prefix=base_prefix,
        tail_prefix=tail_prefix,
        default_sample_rate=default_sample_rate,
        sample_rate_factor=sample_rate_factor,
        emit_legacy=emit_legacy,
        logger=logger)


def get_labeled_statsd_client(conf=None, logger=None):
    """
    Get an instance of LabeledStatsdClient using config settings.

    **config and defaults**::

        log_statsd_host = (disabled)
        log_statsd_port = 8125
        log_statsd_default_sample_rate = 1.0
        log_statsd_sample_rate_factor = 1.0
        statsd_label_mode = disabled

    :param conf: Configuration dict to read settings from
    :param logger: stdlib logger instance used by statsd client for logging
    :return: an instance of ``LabeledStatsdClient``
    """
    conf = conf or {}

    host = conf.get('log_statsd_host')
    port = int(conf.get('log_statsd_port', 8125))
    default_sample_rate = float(
        conf.get('log_statsd_default_sample_rate', 1))
    sample_rate_factor = float(
        conf.get('log_statsd_sample_rate_factor', 1))

    label_mode = conf.get(
        'statsd_label_mode', 'disabled').lower()

    default_labels = {}
    for k, v in conf.items():
        if not k.startswith(STATSD_CONF_USER_LABEL_PREFIX):
            continue
        conf_label = k[len(STATSD_CONF_USER_LABEL_PREFIX):]
        result = USER_LABEL_PATTERN.search(conf_label)
        if result is not None:
            raise ValueError(
                'invalid character in statsd user label '
                'configuration {0!r}: {1!r}'.format(
                    k, result.group(0)))
        result = USER_VALUE_PATTERN.search(v)
        if result is not None:
            raise ValueError(
                'invalid character in configuration {0!r} '
                'value {1!r}: {2!r}'.format(
                    k, v, result.group(0)))
        conf_label = STATSD_USER_LABEL_NAMESPACE + conf_label
        default_labels[conf_label] = v

    return LabeledStatsdClient(
        host, port,
        default_sample_rate=default_sample_rate,
        sample_rate_factor=sample_rate_factor,
        label_mode=label_mode,
        default_labels=default_labels,
        logger=logger)


class AbstractStatsdClient:
    """
    Base class to facilitate sending metrics to a socket. Sub-classes are
    responsible for formatting metrics lines.

    :param host: Statsd host name. If ``None`` then metrics are not sent.
    :param port: Statsd host port.
    :param default_sample_rate: The default rate at which metrics should be
        sampled if no sample rate is otherwise specified. Should be a float
        value between 0 and 1.
    :param sample_rate_factor: A multiplier to apply to the rate at which
        metrics are sampled. Should be a float value between 0 and 1.
    :param logger: A stdlib logger instance.
    """
    def __init__(self, host, port, default_sample_rate=1,
                 sample_rate_factor=1, logger=None):
        self._host = host
        self._port = port
        self._default_sample_rate = default_sample_rate
        self._sample_rate_factor = sample_rate_factor
        self.random = random
        self.logger = logger
        self._sock_family = self._target = None

        if self._host:
            self._set_sock_family_and_target(self._host, self._port)

    def _set_sock_family_and_target(self, host, port):
        # Determine if host is IPv4 or IPv6
        addr_info = None
        try:
            addr_info = socket.getaddrinfo(host, port, socket.AF_INET)
            self._sock_family = socket.AF_INET
        except socket.gaierror:
            try:
                addr_info = socket.getaddrinfo(host, port, socket.AF_INET6)
                self._sock_family = socket.AF_INET6
            except socket.gaierror:
                # Don't keep the server from starting from what could be a
                # transient DNS failure.  Any hostname will get re-resolved as
                # necessary in the .sendto() calls.
                # However, we don't know if we're IPv4 or IPv6 in this case, so
                # we assume legacy IPv4.
                self._sock_family = socket.AF_INET

        # NOTE: we use the original host value, not the DNS-resolved one
        # because if host is a hostname, we don't want to cache the DNS
        # resolution for the entire lifetime of this process.  Let standard
        # name resolution caching take effect.  This should help operators use
        # DNS trickery if they want.
        if addr_info is not None:
            # addr_info is a list of 5-tuples with the following structure:
            #     (family, socktype, proto, canonname, sockaddr)
            # where sockaddr is the only thing of interest to us, and we only
            # use the first result.  We want to use the originally supplied
            # host (see note above) and the remainder of the variable-length
            # sockaddr: IPv4 has (address, port) while IPv6 has (address,
            # port, flow info, scope id).
            sockaddr = addr_info[0][-1]
            self._target = (host,) + (sockaddr[1:])
        else:
            self._target = (host, port)

    def _is_emitted(self, sample_rate):
        """
        Adjust the given ``sample_rate`` by the configured
        ``sample_rate_factor`` and, based on the adjusted sample rate,
        determine if a stat should be emitted on this occasion.

        Sub-classes should call this method before sending a metric line with
        ``_send_line``.

        :param sample_rate: The sample_rate given in the call to emit a stat.
            If ``None`` then this will default to the configured
            ``default_sample_rate``.
        :returns: a tuple ``(<boolean>, <adjusted_sample_rate>)``. The boolean
            is ``True`` if a stat should be emitted on this occasion, ``False``
            otherwise.
        """
        if not self._host:
            # StatsD not configured
            return False, None

        if sample_rate is None:
            sample_rate = self._default_sample_rate
        adjusted_sample_rate = sample_rate * self._sample_rate_factor

        if adjusted_sample_rate < 1 and self.random() >= adjusted_sample_rate:
            return False, None

        return True, adjusted_sample_rate

    def _send_line(self, line):
        """
        Send a ``line`` of metrics to socket.

        Sub-classes should call ``_is_emitted`` before calling this method.

        :param line: The string to be sent to the socket. If ``None`` then
            nothing is sent.
        """

        if line is None:
            return

        # Ideally, we'd cache a sending socket in self, but that
        # results in a socket getting shared by multiple green threads.
        with closing(self._open_socket()) as sock:
            try:
                return sock.sendto(line.encode('utf-8'), self._target)
            except IOError as err:
                if self.logger:
                    self.logger.warning(
                        'Error sending UDP message to %(target)r: %(err)s',
                        {'target': self._target, 'err': err})

    def _open_socket(self):
        return socket.socket(self._sock_family, socket.SOCK_DGRAM)

    def _update_stats(self, metric, value, **kwargs):
        # This method was added to disaggregate *crement metrics when testing
        return self._send(metric, value, 'c', **kwargs)

    def update_stats(self, metric, value, **kwargs):
        self._update_stats(metric, value, **kwargs)

    def increment(self, metric, **kwargs):
        return self._update_stats(metric, 1, **kwargs)

    def decrement(self, metric, **kwargs):
        return self._update_stats(metric, -1, **kwargs)

    def _timing(self, metric, timing_ms, **kwargs):
        # This method was added to disaggregate timing metrics when testing
        return self._send(metric, round(timing_ms, 4), 'ms', **kwargs)

    def timing(self, metric, timing_ms, **kwargs):
        return self._timing(metric, timing_ms, **kwargs)

    def timing_since(self, metric, orig_time, **kwargs):
        return self._timing(
            metric, (time.time() - orig_time) * 1000, **kwargs)

    def transfer_rate(self, metric, elapsed_time, byte_xfer, **kwargs):
        if byte_xfer:
            return self._timing(
                metric, elapsed_time * 1000 / byte_xfer * 1000, **kwargs)


class StatsdClient(AbstractStatsdClient):
    """
    A legacy statsd client.  This client does not support labeled metrics.

    A prefix may be specified using the ``base_prefix`` and ``tail_prefix``
    arguments. The prefix is added to the name of every metric such that the
    emitted metric name has the form:

        [<base_prefix>.][tail_prefix.]<metric name>

    :param host: Statsd host name. If ``None`` then metrics are not sent.
    :param port: Statsd host port.
    :param base_prefix: (optional) A string that will form the first part of a
        prefix added to each metric name. The prefix is separated from the
        metric name by a '.' character.
    :param tail_prefix: (optional) A string that will form the second part of a
        prefix added to each metric name. The prefix is separated from the
        metric name by a '.' character.
    :param default_sample_rate: The default rate at which metrics should be
        sampled if no sample rate is otherwise specified. Should be a float
        value between 0 and 1.
    :param sample_rate_factor: A multiplier to apply to the rate at which
        metrics are sampled. Should be a float value between 0 and 1.
    :param emit_legacy: if ``True`` then the client will emit metrics; if
        ``False``  then the client will emit no metrics.
    :param logger: A stdlib logger instance.
    """
    def __init__(self, host, port,
                 base_prefix='', tail_prefix='',
                 default_sample_rate=1, sample_rate_factor=1,
                 emit_legacy=True,
                 logger=None):
        super().__init__(
            host, port, default_sample_rate, sample_rate_factor, logger)
        self._base_prefix = base_prefix

        self.emit_legacy = emit_legacy

        self._set_prefix(tail_prefix)

    def _send(self, metric, value, metric_type, sample_rate=None):
        is_emitted, adjusted_sample_rate = self._is_emitted(sample_rate)
        if self.emit_legacy and is_emitted:
            metric = self._prefix + metric
            line = _build_line_parts(
                metric, value, metric_type, adjusted_sample_rate)
            return self._send_line(line)

    def _set_prefix(self, tail_prefix):
        """
        Modifies the prefix that is added to metric names. The resulting prefix
        is the concatenation of the component parts `base_prefix` and
        `tail_prefix`. Only truthy components are included. Each included
        component is followed by a period, e.g.::

            <base_prefix>.<tail_prefix>.
            <tail_prefix>.
            <base_prefix>.
            <the empty string>

        Note: this method is expected to be called from the constructor only,
        but exists to provide backwards compatible functionality for the
        deprecated set_prefix() method.

        :param tail_prefix: The new value of tail_prefix
        """
        if tail_prefix and self._base_prefix:
            self._prefix = '.'.join([self._base_prefix, tail_prefix, ''])
        elif tail_prefix:
            self._prefix = tail_prefix + '.'
        elif self._base_prefix:
            self._prefix = self._base_prefix + '.'
        else:
            self._prefix = ''

    def set_prefix(self, tail_prefix):
        """
        This method is deprecated; use the ``tail_prefix`` argument of the
        constructor when instantiating the class instead.
        """
        warnings.warn(
            'set_prefix() is deprecated; use the ``tail_prefix`` argument of '
            'the constructor when instantiating the class instead.',
            DeprecationWarning, stacklevel=2
        )
        self._set_prefix(tail_prefix)

    # for backwards compat StatsdClient supports sample_rate as positional arg
    def update_stats(self, metric, value, sample_rate=None):
        """
        Update a counter, aggregated metric changed by value.

        :param metric: name of the metric
        :param value: int, the counter delta
        :param sample_rate: float, override default sample_rate
        """
        return super().update_stats(metric, value,
                                    sample_rate=sample_rate)

    def increment(self, metric, sample_rate=None):
        """
        Increment a counter, aggregated metric increased by one.

        :param metric: name of the metric
        :param sample_rate: float, override default sample_rate
        """
        return super().increment(metric, sample_rate=sample_rate)

    def decrement(self, metric, sample_rate=None):
        """
        Decrement a counter, aggregated metric decreased by one.

        :param metric: name of the metric
        :param sample_rate: float, override default sample_rate
        """
        return super().decrement(metric, sample_rate=sample_rate)

    def timing(self, metric, timing_ms, sample_rate=None):
        """
        Update a timing metric, aggregated percentiles recalculated.

        :param metric: name of the metric
        :param timing_ms: float, total timing of operation
        :param sample_rate: float, override default sample_rate
        """
        return super().timing(metric, timing_ms, sample_rate=sample_rate)

    def timing_since(self, metric, orig_time, sample_rate=None):
        """
        Update a timing metric, aggregated percentiles recalculated.

        This is an alternative spelling of timing which calculates
        timing_ms=(time.time() - orig_time) for you.

        :param metric: name of the metric
        :param orig_time: float, time.time() from start of operation
        :param sample_rate: float, override default sample_rate
        """
        return super().timing_since(metric, orig_time,
                                    sample_rate=sample_rate)

    def transfer_rate(self, metric, elapsed_time, byte_xfer,
                      sample_rate=None):
        """
        Update a timing metric, aggregated percentiles recalculated.

        This is a timing metric, but adjusts the timing data per kB transferred
        (ms/kB) for each non-zero-byte update.  Allegedly this could be used to
        monitor problematic devices, where higher is bad.

        :param metric: name of the metric
        :param elapsed_time: float, total timing of operation
        :param byte_xfer: int, number of bytes
        :param sample_rate: float, override default sample_rate
        """
        return super().transfer_rate(metric, elapsed_time, byte_xfer,
                                     sample_rate=sample_rate)


class LabeledStatsdClient(AbstractStatsdClient):
    """
    A statsd client that supports annotating metrics with labels.

    Labeled metrics can be emitted in the style of Graphite, Librato, InfluxDB,
    or DogStatsD, by specifying the corresponding ``label_mode`` when
    constructing a client. If ``label_mode`` is ``disabled`` then no metrics
    are emitted by the client.

    Label keys should contain only ASCII letters ('a-z', 'A-Z'), digits
    ('0-9') and the underscore character ('_'). Label values may also contain
    the period ('.') character.

    Callers should avoid using labels that have a high cardinality of values
    since this may result in an unreasonable number of distinct time series for
    collectors to maintain. For example, labels should NOT be used for object
    names or transaction ids.

    :param host: Statsd host name. If ``None`` then metrics are not sent.
    :param port: Statsd host port.
    :param default_sample_rate: The default rate at which metrics should be
        sampled if no sample rate is otherwise specified. Should be a float
        value between 0 and 1.
    :param sample_rate_factor: A multiplier to apply to the rate at which
        metrics are sampled. Should be a float value between 0 and 1.
    :param label_mode: one of 'graphite', 'dogstatsd', 'librato', 'influxdb'
        or 'disabled'.
    :param default_labels: a dictionary of labels that will be added to every
        metric emitted by the client.
    :param logger: A stdlib logger instance.
    """
    def __init__(self, host, port,
                 default_sample_rate=1, sample_rate_factor=1,
                 label_mode='disabled', default_labels=None,
                 logger=None):
        super().__init__(
            host, port, default_sample_rate, sample_rate_factor, logger)
        self.default_labels = default_labels or {}
        self.label_formatter = _get_labeled_statsd_formatter(label_mode)
        if self.logger:
            self.logger.debug('Labeled statsd mode: %s (%s)',
                              label_mode, self.logger.name)

    def _send(self, metric, value, metric_type, labels=None, sample_rate=None):
        if not self.label_formatter:
            return

        is_emitted, adjusted_sample_rate = self._is_emitted(sample_rate)
        if is_emitted:
            return self._send_line(self._build_line(
                metric, value, metric_type, labels, adjusted_sample_rate))

    def _build_line(self, metric, value, metric_type, labels, sample_rate):
        all_labels = dict(self.default_labels)
        if labels:
            all_labels.update(labels)
        return self.label_formatter(
            metric,
            value,
            metric_type,
            sample_rate,
            sorted(all_labels.items()))

    def update_stats(self, metric, value, *, labels=None, sample_rate=None):
        """
        Update a counter, aggregated metric changed by value.

        :param metric: name of the metric
        :param value: int, the counter delta
        :param labels: dict, metric labels
        :param sample_rate: float, override default sample_rate
        """
        return super().update_stats(metric, value, labels=labels,
                                    sample_rate=sample_rate)

    def increment(self, metric, *, labels=None, sample_rate=None):
        """
        Increment a counter, aggregated metric increased by one.

        :param metric: name of the metric
        :param labels: dict, metric labels
        :param sample_rate: float, override default sample_rate
        """
        return super().increment(metric, labels=labels,
                                 sample_rate=sample_rate)

    def decrement(self, metric, *, labels=None, sample_rate=None):
        """
        Decrement a counter, aggregated metric decreased by one.

        :param metric: name of the metric
        :param labels: dict, metric labels
        :param sample_rate: float, override default sample_rate
        """
        return super().decrement(metric, labels=labels,
                                 sample_rate=sample_rate)

    def timing(self, metric, timing_ms, *, labels=None, sample_rate=None):
        """
        Update a timing metric, aggregated percentiles recalculated.

        :param metric: name of the metric
        :param timing_ms: float, total timing of operation
        :param labels: dict, metric labels
        :param sample_rate: float, override default sample_rate
        """
        return super().timing(metric, timing_ms, labels=labels,
                              sample_rate=sample_rate)

    def timing_since(self, metric, orig_time, *,
                     labels=None, sample_rate=None):
        """
        Update a timing metric, aggregated percentiles recalculated.

        This is an alternative spelling of timing which calculates
        timing_ms=(time.time() - orig_time) for you.

        :param metric: name of the metric
        :param orig_time: float, time.time() from start of operation
        :param labels: dict, metric labels
        :param sample_rate: float, override default sample_rate
        """
        return super().timing_since(metric, orig_time, labels=labels,
                                    sample_rate=sample_rate)

    def transfer_rate(self, metric, elapsed_time, byte_xfer, *,
                      labels=None, sample_rate=None):
        """
        Update a timing metric, aggregated percentiles recalculated.

        This is a timing metric, but adjusts the timing data per kB transferred
        (ms/kB) for each non-zero-byte update.  Allegedly this could be used to
        monitor problematic devices, where higher is bad.

        :param metric: name of the metric
        :param elapsed_time: float, total timing of operation
        :param byte_xfer: int, number of bytes
        :param labels: dict, metric labels
        :param sample_rate: float, override default sample_rate
        """
        return super().transfer_rate(metric, elapsed_time, byte_xfer,
                                     labels=labels,
                                     sample_rate=sample_rate)
